 # 
 # Copyright (C) 2012-2019 Daniel Manrique-Vallier
 # 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or (at
 # your option) any later version.
 # 
 # This program is distributed in the hope that it will be useful, but
 # WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # General Public License for more details.
 # 
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 #

######################################################################
# Functions for creating and manipulating lcm_EdImp objects.
# Depends on functions in: 
# - ArrayUtils.R
# - MCMCenv_refClass.R
# (c) Daniel Manrique-Vallier 2013-2020
######################################################################

#prototype:
#SEXP R_Create_Lcm_EdImp2(
#		//data = list(levelsJ=numeric(J), x = array(x, dims=(n, J)), ZeroMC_IJ_original=array(dims=c(number of zeros,J))) 
#		SEXP data_list, 
#		SEXP K, SEXP Nmis_max, SEXP a_alpha, SEXP b_alpha, SEXP a_error, SEXP b_error, 
#		SEXP mask_type,	SEXP len_buffer, SEXP subsamp){
#	//mask_type: 0: No error; 1: all of them; 2: Only inconsistent; 3) Only consistency columns in inconsistent rows.

lcm_EdImp_generator <- setRefClass(
  Class = "lcm_EdImp",
  fields = list(
	J = 'numeric',
	K = 'numeric',
	n = 'numeric',
	mask_type = 'numeric',
	levels = 'list',
	data = 'data.frame'
    ),
  methods = list(
    initialize =   function(x, ZeroMC, K, Nmis_max, a_alpha, b_alpha, a_error, b_error, 
			mask_type = 1, len_buffer, subsamp){
      #mask_type: 0: No error; 1: all of them; 2: Only inconsistent; 3:only inconsistent rows and columns.
      callSuper()
      .self$J = NCOL(x)
      .self$K = K
      .self$n = NROW(x)
      .self$levels = fn_df_levels(x)
      .self$data = x
  	  .self$mask_type = mask_type
      #data_list <- list(
      #	x = fn_dataframe2num_matrix(x,C_style=F, offset=-1, missing=-1) , 
      #	ZeroMC_IJ_original = fn_dataframe2num_matrix(d=ZeroMC, offset=-1, missing=-1, C_style=F),
	    #levelsJ = fn_df_nlevels(x) 
      #)
      Izeros <- NROW(ZeroMC)
      xM <- fn_dataframe2num_matrix(x,C_style=TRUE, offset=-1, missing=-1)
      ZeroMC_IJ_original <- fn_dataframe2num_matrix(d=ZeroMC, offset=-1, missing=-1, C_style=TRUE)
      tmp <- .Call('R_Create_Lcm_EdImp2', 
                   #data_list, 
                   as.integer(.self$J),
                   as.integer(.self$n),
                   as.integer(Izeros),
                   as.integer(xM),
                   as.integer(ZeroMC_IJ_original),
                   as.integer(fn_df_nlevels(x)),
                   as.integer(K), as.integer(Nmis_max), 
                   as.double(a_alpha), as.double(b_alpha), as.double(a_error), as.double(b_error),
                   as.integer(mask_type),
                   as.integer(len_buffer), as.integer(subsamp)
      )
      .self$pointer = tmp
    },
    Get_Imputed_Dataset = function(){
      #SEXP R_get_imputed_x_Lcm_Reg_MCZ(SEXP p);
      tmp <- .Call('R_get_imputed_x_Lcm_EdImp', .self$pointer) 	
      tmp <- t(matrix(tmp, nrow=.self$J))
      tmp <- fn_numeric_to_factor(tmp, list_levls=.self$levels, cols = 1:(.self$J), offset=-1)
      names(tmp) <- names(.self$levels)
      return(tmp)
    }
  ),
  contains = "MCMCenviron"
)

lcmEditImput <- function(data, edits, K=20, a_alpha = 0.25, b_alpha = 0.25, a_error = 1, b_error = 1, 
                         mask_type = 1, Nmis_max = 100000,
                         burnin_run = 1000, len_buffer = 10000, thining = 10, verbose = TRUE){
  s <- lcm_EdImp_generator$new(x=data,ZeroMC=edits, K=K, Nmis_max=Nmis_max, a_alpha=a_alpha, b_alpha=b_alpha, 
                               a_error = a_error, b_error = b_error,
                               mask_type = mask_type,
                               len_buffer=len_buffer, subsamp=thining)
  s$Init_Model(output = verbose)
  s$Update(burnin_run, output = verbose)
  return(s)
}

edit_impute <- function(object, burnin=1000, samples = 10, thining = 100, verbose = TRUE){
  if (class(object) != 'lcm_EdImp'){
    stop('Argument "object" must be of lcm_EdImp class.')
  }
  object$Update(burnin, output= verbose)
  mi_datasets <- list()
  for (i in 1:samples){
    object$Update(thining, output = verbose)
    mi_datasets[[i]] <- object$Get_Imputed_Dataset()
  }
  return(mi_datasets)
}

fn_partial_contingency <- function(dat, responseName = "Freq"){
  x <- fn_dataframe2num_matrix(dat,C_style=T, offset=-1, missing=-1)
  lev <- fn_df_nlevels(dat)
  #SEXP R_Partial_Contingency_Table(SEXP dataIJ_flat, SEXP levelsJ);
  tmp <- .Call('R_Partial_Contingency_Table', as.integer(x), as.integer(lev))
  r <- tmp[[1]]
  r <- matrix(r, byrow=T, ncol = NCOL(dat))
  r <- fn_numeric_to_factor(r, list_levls=fn_df_levels(dat), missing=-1, offset=-1)
  r <- data.frame(r, tmp[[2]])
  names(r) <- c(names(dat),responseName)
  return(r)
} 

lcm_Synth_generator <- setRefClass(
  Class = "lcm_Synth",
  fields = list(
	J = 'numeric',
	K = 'numeric',
	n = 'numeric',
	nSynth = 'numeric',
	levels = 'list',
	data = 'data.frame'
    ),
  methods = list(
    initialize =   function(x, ZeroMC, K, a_alpha, b_alpha, N_Synth, method = 'rejection', len_buffer, subsamp){
      callSuper()
      .self$J = NCOL(x)
      .self$K = K
      .self$n = NROW(x)
      .self$levels = fn_df_levels(x)
      .self$data = x
	  .self$n <- NROW(x)
	meth <- switch(method, rejection = 0, gibbs = 1, metropolis1 = 2, chain = 3)
	if (length(meth) == 0) stop('Unrecognized imputation method')
	#SEXP R_Create_Lcm_Synth(SEXP x_flat, SEXP J, SEXP n, SEXP ZeroMC_flat, SEXP nZeroMC, SEXP levels,
	#					  SEXP K, SEXP Nmis_max, SEXP a_alpha, SEXP b_alpha,
	#					  SEXP N_Synth, SEXP method
	#					  SEXP len_buffer, SEXP subsamp)
      #Allow for the possibility of no zeros.
	  if (!is.null(ZeroMC)){
		ZEROS <- fn_dataframe2num_matrix(d=ZeroMC, offset=-1, missing=-1, C_style=T)
		NZEROS <- NROW(ZeroMC)
	  } else {
		ZEROS <- 0
		NZEROS <- 0
	  }
	  tmp <- .Call('R_Create_Lcm_Synth', 
					as.integer(fn_dataframe2num_matrix(d=x, offset=-1, missing=-1, C_style=T)), #SEXP x_flat, 
					as.integer(.self$J), #SEXP J, 
					as.integer(.self$n), #SEXP n, 
					as.integer(ZEROS), #SEXP ZeroMC_flat,
					as.integer(NZEROS), #SEXP nZeroMC,
					as.integer(fn_df_nlevels(x)), #SEXP levels,
					as.integer(K), #SEXP K,
				    	#as.integer(Nmis_max), #SEXP Nmis_max, 
					as.double(a_alpha), #SEXP a_alpha, 
					as.double(b_alpha), #SEXP b_alpha,
					as.integer(N_Synth), #SEXP N_Synth,
					as.integer(meth), #SEXP method, //method = 0 -> Rejection / method = 1 -> Gibbs
					as.integer(len_buffer), as.integer(subsamp)
      )
      .self$pointer = tmp
    },
    Get_Imputed_Dataset = function(){
      #SEXP R_get_imputed_x_Lcm_Synth(SEXP p);
      tmp <- .Call('R_get_imputed_x_Lcm_Synth', .self$pointer) 	
      tmp <- t(matrix(tmp, nrow=.self$J))
      tmp <- fn_numeric_to_factor(tmp, list_levls=.self$levels, cols = 1:(.self$J), offset=-1)
      names(tmp) <- names(.self$levels)
      return(tmp)
    },
	Get_Imputed_Dataset_TRUNCATE = function(MU){
	  MUraw <- as.integer(fn_dataframe2num_matrix(d=MU, offset=-1, missing=-1, C_style=T))
	  nMU <- as.integer(NROW(MU))
      #SEXP R_get_imputed_x_Lcm_Synth_TRUNCATE(SEXP p, SEXP MUS, SEXP nMUS);
      tmp <- .Call('R_get_imputed_x_Lcm_Synth', .self$pointer, MUraw, nMU) 	
      tmp <- t(matrix(tmp, nrow=.self$J))
      tmp <- fn_numeric_to_factor(tmp, list_levls=.self$levels, cols = 1:(.self$J), offset=-1)
      names(tmp) <- names(.self$levels)
      return(tmp)
    }#,

#	Resize_Nmis_Max = function(newsize){
#		#R_resize_imput_buffer_Lcm_Synth(SEXP p, SEXP newsize)
#		.Call('R_resize_imput_buffer_Lcm_Synth', .self$pointer, as.integer(newsize))
#	}
  ),
  contains = "MCMCenviron"
)

lcm_Imput_Generator <- setRefClass(
  Class = "lcm_Imput",
  fields = list(
	J = 'numeric',
	K = 'numeric',
	n = 'numeric',
	levels = 'list',
	data = 'data.frame'
    ),
  methods = list(
    initialize =   function(x, ZeroMC, K, Nmis_max, a_alpha, b_alpha, len_buffer, subsamp){
      callSuper()
      .self$J = NCOL(x)
      .self$K = K
      .self$n = NROW(x)
      .self$levels = fn_df_levels(x)
      .self$data = x
	  .self$n <- NROW(x)
	#SEXP R_Create_Lcm_Imput(SEXP x_flat, SEXP J, SEXP n, SEXP ZeroMC_flat, SEXP nZeroMC, SEXP levels,
	#					  SEXP K, SEXP Nmis_max, SEXP a_alpha, SEXP b_alpha,
	#					  SEXP len_buffer, SEXP subsamp)
      tmp <- .Call('R_Create_Lcm_Imput', 
					as.integer(fn_dataframe2num_matrix(d=x, offset=-1, missing=-1, C_style=T)), #SEXP x_flat, 
					as.integer(.self$J), #SEXP J, 
					as.integer(.self$n), #SEXP n, 
					as.integer(fn_dataframe2num_matrix(d=ZeroMC, offset=-1, missing=-1, C_style=T)), #SEXP ZeroMC_flat,
					as.integer(NCOL(ZeroMC)), #SEXP nZeroMC,
					as.integer(fn_df_nlevels(x)), #SEXP levels,
					as.integer(K), #SEXP K,
				    as.integer(Nmis_max), #SEXP Nmis_max, 
					as.double(a_alpha), #SEXP a_alpha, 
					as.double(b_alpha), #SEXP b_alpha,
					as.integer(len_buffer), as.integer(subsamp)
      )
      .self$pointer = tmp
    },
    Get_Imputed_Dataset = function(){
      #SEXP R_get_imputed_x_Lcm_Imput(SEXP p);
      tmp <- .Call('R_get_imputed_x_Lcm_Imput', .self$pointer) 	
      tmp <- t(matrix(tmp, nrow=.self$J))
      tmp <- fn_numeric_to_factor(tmp, list_levls=.self$levels, cols = 1:(.self$J), offset=-1)
      names(tmp) <- names(.self$levels)
      return(tmp)
    }
  ),
  contains = "MCMCenviron"
)


#.onUnload <- function (libpath) {
#  gc(verbose = FALSE)
#  library.dynam.unload("lcmEdImp", libpath)
#}
