/* 
 * Copyright (C) 2007-2019 Daniel Manrique-Vallier
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf.h> //special functions (for gamma)
#include <string.h>
#include <math.h>
#include "daniel2/dan_math_gsl.h"
#include "CLcm_Bay.h"

/*------------------------------------------------------------
Implementation of class CParams_Lcm_Bay 

------------------------------------------------------------*/
//Generic Constructor
void CParams_Lcm_Bay::class_construct(int _J, int _K, int *_levelsJ, int _n){
	//init members	
	this->J=_J; this->K=_K; this->n = _n; 
	this->levelsJ = _levelsJ;
	this->L = *std::max_element(_levelsJ, _levelsJ + _J);//dan_max_vector_int(levelsJ, J);
	//allocate
	_add_parameter("psiJKL", CPar_Data_Type::T_DOUBLE, &psiJKL, 3, _J, _K, L);
	_add_parameter("zI",  CPar_Data_Type::T_INT, &zI, 1, _n);
	_add_parameter("nuK", CPar_Data_Type::T_DOUBLE, &nuK, 1, _K);
	_add_parameter("count0K", CPar_Data_Type::T_INT, &count0K, 1, _K);
	_add_parameter("count1K", CPar_Data_Type::T_INT, &count1K, 1, _K);
	_add_parameter("aux_dirJKL", CPar_Data_Type::T_INT, &aux_dirJKL, 3, _J, _K, L);
	_add_existing_scalar("K", CPar_Data_Type::T_INT, &K);
	(*this)["psiJKL"].fill(0.0);
}

void CParams_Lcm_Bay::initizalize(gsl_rng *r){
	//Initializes the parameters for the chain.
	int j,k,l;
	for (k =0; k < K; k++){
		this->nuK[k] = 1.0 / double(K);
	}

	for(j = 0; j < J; j++){
		for(k = 0; k < K; k++){
			for(l = 0; l < this->levelsJ[j]; l++){
				psiJKL[j][k][l] = 1.0 / double(levelsJ[j]);
			}
		}
	}
	//initialize missing data.
}



//--------------------------------------------------------------------------------
// Implementation of class CLcm_Bay
//--------------------------------------------------------------------------------
void CLcm_Bay::Update() {
	sam_z();
	sam_psi();
	sam_nu();
}
inline
void CLcm_Bay::sam_psi(){
	static double dirL[1000];
	int i, j, k, l;
	memset(par->aux_dirJKL[0][0], 0, sizeof(int) * data->J * par->K * data->L);
	for (i = 0; i < data->n; i ++){
		for (j = 0; j < par->J; j++){
			k = par->zI[i];
			l = data->x[i][j];
			par->aux_dirJKL[j][k][l]++;
		}
	}
	//add the prior (+1) and sample
	for(j = 0; j < data->J; j++){
		for(k = 0; k < par->K; k++){
			//add prior and cast.
			for(l = 0; l < data->levelsJ[j]; l++){
				par->aux_dirJKL[j][k][l]++; //prior
				dirL[l] = double(par->aux_dirJKL[j][k][l]);
			}
			//sample!
			gsl_ran_dirichlet(r, data->levelsJ[j], dirL, par->psiJKL[j][k]);
		}
	}
}

inline
void CLcm_Bay::sam_z(){
	//Here we also count how many individuals belong to each class.
	int i, j, k;
	double prod;
	static double probs[1000];
	
	//memset(par->countK, 0, sizeof(int)*par->K);
	std::fill(par->count1K, par->count1K + par->K, 0);
	for (i = 0; i < data->n; i++){
		for (k = 0; k < par->K; k++){
			prod= 1.0;
			for (j = 0; j < par->J; j++){
				prod *= par->psiJKL[j][k][ data->x[i][j] ];
			}
			probs[k] = prod * par->nuK[k];
		}
		//sample zi
		par->zI[i] = k = dan_multinomial_1(r, par->K, probs, false);
		par->count1K[k]++;
	}
}

void CLcm_Bay::sam_nu(){
	double acc = 0.0;
	double g;
	int k;
	for(k = 0; k < par->K; k++){
		g = gsl_ran_gamma(r, 1.0 + double(par->count1K[k]), 1.0);
		par->nuK[k] = g;
		acc += g;
	}
	for(k = 0; k < par->K; k++){
		par->nuK[k] /= acc;
	}
}

double CLcm_Bay::predict(int* xJ){
	int j,k;
	double prod;
	double sum = 0.0;
	for(k = 0; k < par->K; k ++){
		prod = 1.0;
		for(j = 0; j < par->J; j++){
			prod *= par->psiJKL[j][k][ xJ[j] ];
		}
		sum += par->nuK[k] * prod;
	}
	return(sum);
}
