/* 
 * Copyright (C) 2007-2019 Daniel Manrique-Vallier
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf.h> //special functions (for gamma)
#include <string.h>
#include <math.h>
#include "CLcm_Bay_Reg.h"
#include "daniel2/dan_math_gsl.h"

/*------------------------------------------------------------
Implementation of class CParams_Lcm_Bay_Reg 

------------------------------------------------------------*/
void CParams_Lcm_Bay_Reg::initizalize(gsl_rng *r){
	CParams_Lcm_Bay::initizalize(r);
	//just the extra.
	storage["log_nuK"].fill(-log(double(K))); //.fill("log_nuK", -log(double(K)));
	alpha = 1 ; //initialize at full spread: uniform prior.
}



//--------------------------------------------------------------------------------
// Implementation of class CLcm_Bay_Reg
//--------------------------------------------------------------------------------
void CLcm_Bay_Reg::Update() {
	sam_z();
	sam_psi();
	sam_nu();
	sam_alpha();
	par->compute_k_star();
}
inline
void CLcm_Bay_Reg::sam_psi(){
	static double dirL[1000];
	int i, j, k, l;
	memset(par->aux_dirJKL[0][0], 0, sizeof(int) * data->J * par->K * data->L);
	for (i = 0; i < data->n; i ++){
		for (j = 0; j < par->J; j++){
			k = par->zI[i];
			l = data->x[i][j];
			par->aux_dirJKL[j][k][l]++;
		}
	}
	//add the prior (+1) and sample
	for(j = 0; j < data->J; j++){
		for(k = 0; k < par->K; k++){
			//add prior and cast.
			for(l = 0; l < data->levelsJ[j]; l++){
				par->aux_dirJKL[j][k][l]++; //prior
				dirL[l] = double(par->aux_dirJKL[j][k][l]);
			}
			//sample!
			gsl_ran_dirichlet(r, data->levelsJ[j], dirL, par->psiJKL[j][k]);
		}
	}
}

inline
void CLcm_Bay_Reg::sam_z(){
	//Here we also count how many individuals belong to each class.
	int i, j, k;
	double prod;
	static double probs[1000];
	
	memset(par->count1K, 0, sizeof(int)*par->K);
	for (i = 0; i < data->n; i++){
		for (k = 0; k < par->K; k++){
			prod= 1.0;
			for (j = 0; j < par->J; j++){
				prod *= par->psiJKL[j][k][ data->x[i][j] ];
			}
			probs[k] = prod * par->nuK[k];
		}
		//sample zi
		par->zI[i] = k = dan_multinomial_1(r, par->K, probs, false);
		par->count1K[k]++;
	}
}
//inline
//void CLcm_Bay_Reg::sam_nu(){
//	double acc = 0.0;
//	double g;
//	int k;
//	for(k = 0; k < par->K; k++){
//		g = gsl_ran_gamma(r, par->alpha + double(par->countK[k]), 1.0); //prior = alpha * (1,1,1,1,1...,1)
//		par->nuK[k] = g;
//		acc += g;
//	}
//	for(k = 0; k < par->K; k++){
//		par->nuK[k] /= acc;
//		par->log_nuK[k] = log(par->nuK[k]);
//	}
//}
//
//inline
//void CLcm_Bay_Reg::sam_alpha(){
//	double dK = double(par->K);
//	//metropolis hastings step.
//	double sum_log_nu = dan_sum_vec(par->log_nuK, par->K);
//	double alpha_star = gsl_ran_gamma(r, par->a_alpha + dK - 1.0, 1.0 / (par->b_alpha - sum_log_nu));
//	double lratio = dan_lratio_gamma(dK * alpha_star, dK * par->alpha);
//	lratio += dK *(dan_lratio_gamma(par->alpha, alpha_star));
//	lratio += (par->a_alpha - 1.0) * (log(alpha_star) - log(par->alpha));
//	lratio += (alpha_star - par->alpha) * (-par->b_alpha + sum_log_nu);
//	//acceptance step
//	if (lratio>=0) goto accept;
//	else if (log(gsl_rng_uniform_pos(r)) <= lratio) goto accept;
//	return;
//accept:
//	par->alpha = alpha_star;
//}
inline
void CLcm_Bay_Reg::sam_nu(){
	int k;
	double b = 0.0, a = 0.0;
	int n_acc = 0;
	//double acc_prob = 0.0;
	double l_acc_prob = -DAN_INF;
	double l_acc_prod = 0.0;
	double V;//, one_minus_V;
	for (k = 0; k < par->K - 1; ++k){
		n_acc += par->count1K[k];
		a = double(1 + par->count1K[k]);
		b = par->alpha + double(data->n - n_acc);
		V = gsl_ran_beta(r, a, b);
		par->log_nuK[k] = (log(V) + l_acc_prod);
		if (par->log_nuK[k] < -50) par->log_nuK[k] = -50;//approx 1e-22
		l_acc_prod += log(1.0 - V);
		l_acc_prob = dan_log_sum(l_acc_prob, par->log_nuK[k]);
		par->nuK[k] = exp(par->log_nuK[k]);
	}
	par->log_nuK[par->K -1] = l_acc_prod;
	if (par->log_nuK[par->K-1] < -50) par->log_nuK[par->K - 1] = -50;//approx 1e-44
	par->nuK[par->K - 1] = exp(par->log_nuK[par->K-1]);
}

inline
void CLcm_Bay_Reg::sam_alpha(){
	par->alpha = gsl_ran_gamma(r, par->a_alpha + double(par->K) - 1.0, 1.0 / (par->b_alpha - par->log_nuK[par->K-1]));
}
double CLcm_Bay_Reg::predict(int* xJ){
	int j,k;
	double prod;
	double sum = 0.0;
	for(k = 0; k < par->K; k ++){
		prod = 1.0;
		for(j = 0; j < par->J; j++){
			prod *= par->psiJKL[j][k][ xJ[j] ];
		}
		sum += par->nuK[k] * prod;
	}
	return(sum);
}

