/* 
 * Copyright (C) 2007-2019 Daniel Manrique-Vallier
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <string.h>
#include <math.h>
#include <queue>
#include <numeric>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf.h> //special functions (for gamma)

#include "daniel2/dan_math_gsl.h"
#include "CData_DM_MCZ.h"
#include "CLcm_Bay_Reg_MCZ.h"
#include "margin_conditions.h"
#include "definitions.h"

//--------------------------------------------------------------------------------
// Implementation of class CParams_Lcm_EdImp
//--------------------------------------------------------------------------------
void CParams_Lcm_EdImp::class_construct(int _Nmis_max, int** _MCZ, int _nZeroMC, int **_x, int **_emask){
	//this assumes that base class has been initialized
	//initialize members.
	//this->Nmis = _Nmis;
	this->N_mis_max = _Nmis_max;
	this->nZeroMC = _nZeroMC;

	//allocate extra.
	_add_parameter("MCZ", CPar_Data_Type::T_INT, &MCZ, 2, _nZeroMC, J); 
	_add_parameter("xIJ", CPar_Data_Type::T_INT, &xIJ, 2, n, J); 
	_add_parameter("eIJ", CPar_Data_Type::T_INT, &eIJ, 2, n, J);
	_add_parameter("eMaskIJ", CPar_Data_Type::T_INT, &eMaskIJ, 2, n, J);
	//_add_parameter("e_marginal_J", CPar_Data_Type::T_INT, &e_marginal_J, 1, J);
	_add_existing_scalar("e_marginal", CPar_Data_Type::T_INT, &e_marginal);
	//_add_parameter("epsilonJ", CPar_Data_Type::T_DOUBLE, &epsilonJ, 1, J);
	_add_existing_scalar("epsilon", CPar_Data_Type::T_DOUBLE, &epsilon);
	_add_parameter("qJ", CPar_Data_Type::T_DOUBLE, &qJ, 1, J);
	_add_parameter("pZeroMC_I", CPar_Data_Type::T_DOUBLE, &pZeroMC_I, 1, _nZeroMC);
	_add_parameter("z2_Nmax", CPar_Data_Type::T_INT, &z2_Nmax, 1, _Nmis_max);
	_add_parameter("x2_NMax_J", CPar_Data_Type::T_INT, &x2_NMax_J, 2, _Nmis_max, J);
	_add_parameter("count_partition", CPar_Data_Type::T_INT, &count_partition, 1, _nZeroMC);
	_add_parameter("permutationJ", CPar_Data_Type::T_INT, &permutationJ, 1, J);
	//register existing.
	_add_existing_scalar("prob_zero", CPar_Data_Type::T_DOUBLE, &prob_zero);
	_add_existing_scalar("Nmis", CPar_Data_Type::T_INT, &Nmis);
	_add_existing_scalar("N_mis_max", CPar_Data_Type::T_INT, &N_mis_max);
	_add_existing_scalar("nZeroMC", CPar_Data_Type::T_INT, &nZeroMC);
	//fill x with data
	(*this)["xIJ"].copy_from_raw(_x[0]); //NOTE THAT HERE X INCLUDES INCONSISTENT VALUES!!! HAVE TO INITIALIZE.
	//fill eMask
	(*this)["eMaskIJ"].copy_from_raw(_emask[0]);
	//copy MCZ
	(*this)["MCZ"].copy_from_raw(_MCZ[0]);
	//initial permutation
	for (int j = 0; j < J; j++) permutationJ[j] = j;
	//temporary storage:
	_add_parameter("tmp_doubleJ", CPar_Data_Type::T_DOUBLE, &tmp_doubleJ, 1, J);
	_add_parameter("tmp_intJ", CPar_Data_Type::T_INT, &tmp_intJ, 1, J);
	//Synthetic data!
	//_add_parameter("x_synthIJ", CPar_Data_Type::T_INT, &x_synthIJ, 2, n, J);
}
bool CParams_Lcm_EdImp::ValidateMask(int** data_xIJ){
	//verify if the "known" values are consistent.
	//
	for (int i = 0; i < n; i++){
		for (int j=0; j < J; j++){
			if (eMaskIJ[i][j] == 0 && (data_xIJ[i][j] != xIJ[i][j])) return false;	
		}
	}
	return true;
}

void CParams_Lcm_EdImp::initizalize(gsl_rng *r){
	std::vector<double> p(this->L); //reserve at least the maximum number of levels.
	std::fill(p.begin(), p.end(), 1.0); //this is a vector of unnormalized uniform probabilities.
	CParams_Lcm_Bay_Reg::initizalize(r); //parent class
	prob_zero = 0.0;
	Nmis = 0;
	for (int i = 0; i < this->nZeroMC; i++){
		count_partition[i] = 0;
	}
	//memset(this->countK, 0, sizeof(int) * this->K);
	(*this)["count0K"].fill<int>(0);
	(*this)["count1K"].fill<int>(0);
	//std::fill(e_marginal_J, e_marginal_J + J, 0);
	e_marginal = 0;
	for (int j = 0; j < J; j++){
		qJ[j] = 1.0 / double(levelsJ[j] - 1);
		epsilon = 0.05; //kind of small;
	}
	//copy emask to eIJ:
	(*this)["eIJ"].copy_from_raw(eMaskIJ[0]);
	//In eIJ We can only modify those values that are -1. In X we can only modify those where the mask is either -1 or 1 (not known if error / error).

	//Initialize values for xIJ.
	//determine which coordinates (j) determine the MCZ set (just to blank the right values in par->xIJ.)
	std::set<int> MCZcoordinates;
	for (int ii = 0; ii < nZeroMC; ii++){
		for (int j = 0; j < J; j++){
			if (MCZ[ii][j] != -1) MCZcoordinates.insert(j);
		}
	}
	
	bool badvalue = false, inconsistent = false;
	std::vector<int> dubious; //list of indexes of coordinates where the values might be in error (info from the MASK)
	for(int i = 0; i < n; ++i){
		inconsistent = false;
		std::vector<int> x_working(xIJ[i], xIJ[i]+J);
		//Get the coordinates where data might be in error (from the error mask)
		dubious.clear();
		for (int j = 0; j < J; j++) {
			if (eMaskIJ[i][j] == -1) {
				dubious.push_back(j);
			}
		}
		if (dubious.size() == 0) continue; //IF THE MASK DOESN'T QUALIFY ANY OF THE VALUES AS WITH UNKONWN ERRORS, MOVE TO THE NEXT ONE! 
		//Determine if the data has inconsistent values.
		for (int ii  = 0; ii < this->nZeroMC; ++ii){
			if (!check_x_notin_mu(x_working.begin(), x_working.end(), MCZ[ii])) {
				inconsistent = true;
				break; 
			}
		}
		//"Repair" the data: Get initial values for the "true" data that are consistent.
		//<--TODO: CHECK EVERY TIME WE SAMPLE A NEW VALUE (NO NEED TO CHANGE SO MUCH IF X IS ALREADY CONSISTENT!)
		if (inconsistent){
			do{
				//only change those values allowed by the mask. If the mask is wrongly specifed... too bad.
				for (std::vector<int>::iterator it = dubious.begin(); it != dubious.end(); it++){
					//only need to change those that are part of the margin conditions.
					if(MCZcoordinates.find(*it) != MCZcoordinates.end())
						x_working[*it] = dan_multinomial_1(r,  this->levelsJ[*it], &(p[0]), false);
				}
				//check
				badvalue = false;
				for (int ii  = 0; ii < this->nZeroMC; ++ii){
					if (!check_x_notin_mu(x_working.begin(), x_working.end(), MCZ[ii])) {
						badvalue = true; 
						break; 
					}
				}
			} while ( badvalue );
		}
		//Determine which are "in error": 
		// Variable eIJ has to be consistent in indicating which data is distinct from their underlying values. 
		// THIS IS THE LAST CHANCE BECAUSE IN THE NEXT STEP WE LOSE THE ORIGINAL VALUE OF THE DATA.
		// also fill e_marginal_J.
		//Note that since we've only changed the variables that were allowed by the mask, those are the only ones that could be in error.
		for (int j = 0; j < J; j++){
			//e_marginal_J[j] += eIJ[i][j] = (x_working[j] != xIJ[i][j]) ? 1 : 0;
			e_marginal += eIJ[i][j] = (x_working[j] != xIJ[i][j]) ? 1 : 0;
		}
		std::copy<std::vector<int>::iterator, int*>(x_working.begin(), x_working.end(), xIJ[i]);
	}
	//Initialize synthetic data
	//(*this)["x_synthIJ"].copy_from_raw(xIJ[0]);
}

inline
int CParams_Lcm_EdImp::dubious_coordinates(int* xiJ, int* answerJ){
	int ncoord = 0;
	for (int i = 0; i < nZeroMC; i++){
		if( !check_x_notin_mu(xiJ, xiJ + J, MCZ[i]) ){
			for(int j = 0; j < J; j++){
				if (MCZ[i][j] != -1) answerJ[ncoord++] = j;
			}		
			break;
		}
	}
	return ncoord;
}

inline
bool CParams_Lcm_EdImp::is_mask_unknown (const int &i){
	//Determines if any coordinate of the mask is -1
	int *M = eMaskIJ[i];
	for (int j = 0; j < J; j++){
		if (M[j] == -1) return true;
	}
	return false;
}

//--------------------------------------------------------------------------------
// Implementation of class CLcm_EdImp
//--------------------------------------------------------------------------------


void CLcm_EdImp::Update() {
	//regular stuff
	sam_z();
	sam_psi();
	sam_nu();
	//truncation!
	compute_probs_miss();
	sam_Nmis();
	sam_Z2_X2();
	//and regularization!
	sam_alpha();
	// Imputation step
	//sam_e_x();
	sam_e_x_Rejection();
	sam_epsilon();
	//sam_Synthetic();
	par->compute_k_star();
}

// Samplers for the full conditional distributions.
//  each function update the values on the *par object directly.


void CLcm_EdImp::sam_epsilon(){
	//for (int j = 0; j < par->J; j++){
	//	par->epsilonJ[j] = gsl_ran_beta(r, par->a_error + double(par->e_marginal_J[j]), par->b_error + double(data->n - par->e_marginal_J[j]));
	//}
	par->epsilon = gsl_ran_beta(r, par->a_error + double(par->e_marginal), par->b_error + double(data->n * par->J - par->e_marginal));
}

inline
void CLcm_EdImp::sam_e_x_Rejection(){
	//imputation 
	//std::fill(par->e_marginal_J, par->e_marginal_J + par->J, 0); //clear marginal frequcies of error.
	//implementation by rejection sampling
	static double probs[1000];
	//CVariable_Container& psiJKL = (*par)["psiJKL"];
	par->e_marginal = 0;

	for (int i = 0; i < data->n; i ++){
		int* xiJ = par->xIJ[i];
		int z = par->zI[i];
		bool rejected = true;
		int counter = 0;
		do {
			for (int j = 0; j < par->J; j++){
				for (int l = 0; l < par->levelsJ[j]; l++){
					double mult = (l != data->x[i][j]) ?  par->epsilon * par->qJ[j] : 1.0 - par->epsilon; 
					probs[l] = par->psiJKL[j][z][l] * mult;
					//probs[l] = psiJKL.arr_elem<double>(j, z, l) * mult; //this may be faster. Test it!
				}
				xiJ[j] = dan_multinomial_1(r, par->levelsJ[j], probs, false);
			}
			//check if it is a wrong value
			for (int m = 0; m < data->nZeroMC_original; m++){
				rejected = !check_x_notin_mu(xiJ, xiJ + par->J, data->ZeroMC_IJ_original[m]);
				if (rejected) break;
			}
			//if too many attempts, we just give up and sample directly from the
			// actual distribution.
			//[ ]  <<<- TODO: MAKE THE THRESHOLD SETTABLE.
			if (++counter > 1000){
				samp_x_i(i);
				rejected = false;
			}
		} while (rejected);

		for (int j = 0; j < par->J; j++){
			par->e_marginal += par->eIJ[i][j] = (xiJ[j] != data->x[i][j]) ? 1 : 0;
		}
	} //i
}

inline void CLcm_EdImp::samp_x_i(const int& i){
	//sample coordinate by coordinate
	//Updates x_i, *but not* e_{ij}
	int nA;
	int _A[1000];
	static double p[1000];
	static double tmp_prob[1000];
	std::vector<int> running(par->J);

	int* xiJ = par->xIJ[i];
	std::fill(running.begin(), running.end(), -1);
	for (int j = 0; j < par->J; j++){
		nA = 0;
		double normalization = 0.0;
		for (int l = 0; l < par->levelsJ[j]; l++){
			running[j] = l;
			normalization += tmp_prob[l] = _Conditional_X_j(i, j, running);
		}
		for (int l = 0; l < par->levelsJ[j]; l++){
			//get rid of small probs!
			if (1000.0 * double(par->levelsJ[j]) * tmp_prob[l] >= normalization){
				p[nA] = tmp_prob[l];
				_A[nA++] = l;
			}
		}
		running[j] = xiJ[j] = _A[ dan_multinomial_1(r, nA, p, false) ];
		
		//par->e_marginal += par->eIJ[i][j] = (xiJ[j] != data->x[i][j]) ? 1 : 0;
	}
}

inline
void CLcm_EdImp::sam_e_x(){
	//imputation step as gibbs sampler.
	//std::fill(par->e_marginal_J, par->e_marginal_J + par->J, 0); //clear marginal frequcies of error.
	int nA;
	int _A[1000];
	static double p[1000];
	static double tmp_prob[1000];
	std::vector<int> running(par->J);
	par->e_marginal = 0;
	for (int i = 0; i < data->n; i ++){
		//sample coordinate by coordinate
		int* xiJ = par->xIJ[i];
		std::fill(running.begin(), running.end(), -1);
		for (int j = 0; j < par->J; j++){
			nA = 0;
			double normalization = 0.0;
			//for (int l = 0; l < par->levelsJ[j]; l++){
			//	running[j] = l;
			//	double prob = _Conditional_X_j(i, j, running);
			//	if (prob > 1e-20){
			//		//Eliminate small values (rounding errors!)
			//		//write something better. Decide the cutoff in relative terms.
			//		p[nA] = prob;
			//		_A[nA++] = l;
			//	}
			//}
			for (int l = 0; l < par->levelsJ[j]; l++){
				running[j] = l;
				normalization += tmp_prob[l] = _Conditional_X_j(i, j, running);
			}
			for (int l = 0; l < par->levelsJ[j]; l++){
				//get rid of small probs!
				if (1000.0 * double(par->levelsJ[j]) * tmp_prob[l] >= normalization){
					p[nA] = tmp_prob[l];
					_A[nA++] = l;
				}
			}

			running[j] = xiJ[j] = _A[ dan_multinomial_1(r, nA, p, false) ];
			par->e_marginal += par->eIJ[i][j] = (xiJ[j] != data->x[i][j]) ? 1 : 0;
		}
	} //i
}

inline
double CLcm_EdImp::_aj(const int& j, const int& k, const int& xj, const int* y){
	return (xj == y[j]) ? 
		par->psiJKL[j][k][xj] * (1 - par->epsilon) :
		par->psiJKL[j][k][xj] * par->epsilon * par->qJ[j];
}

inline 
double CLcm_EdImp::_Conditional_X_j(const int& i, const int& m, const std::vector<int>& omega_m){
	//Compute the (unnormalized) probability Pr( X_j= x_m) conditional on X_1...X_{m-1}.
	//the argument is a vector xi = (x1, x2,..,xm,-1,-1,-1);
	static int inters[1000];
	static double bJ[100];
	int& z = par->zI[i];
	int* y = data->x[i];
	
	//Compute the b_j's and their product from m+1
	double prod_b = 1.0;
	for (int j = m + 1; j < par->J; j++){
		//note that bJ[j] is undefined for j = 0...m
		bJ[j] = 0.0;
		for (int xj = 0; xj < par->levelsJ[j]; xj++){
			bJ[j] += _aj(j, z, xj, y);	
		}
		prod_b *= bJ[j];
	}
	//intersect with each margin condition
	double sum = 0.0;
	for (int c = 0; c < data->nZeroMC; c++){
		//intersect each margin condition with mc_fixed.
		bool hay_inters = MC_Intersection(omega_m.begin(), omega_m.end(), par->MCZ[c], inters);
		if (!hay_inters) continue;
		double prod = 1.0;
		for (int j = m + 1; j < par->J; j++){
			if (inters[j] == -1) 
				prod *= bJ[j];
			else
				prod *= _aj(j, z, inters[j], y);
		}
		sum += prod;
	}
	return (_aj(m, z, omega_m[m], y) * (prod_b - sum));
}


inline
void CLcm_EdImp::sam_psi(){
	//static double dirL[1000];
	int i, j, k, l;
	std::fill_n(par->aux_dirJKL[0][0], data->J * par->K *data->L, 0);
	//the data part.
	for (i = 0; i < data->n; ++i){
		int k = par->zI[i];
		for (j = 0; j < par->J; ++j){
			++par->aux_dirJKL[j][k][ par->xIJ[i][j] ];
		}
	}
	//the imputed part.
	for (i = 0; i < par->Nmis; ++i){
		k = par->z2_Nmax[i];
		for (j = 0; j < par->J; ++j){
			++par->aux_dirJKL[j][k][ par->x2_NMax_J[i][j] ];
		}
	}
	
	//add the prior (+1) and sample
	for(j = 0; j < data->J; ++j){
		double **psijKL = par->psiJKL[j]; 
		int **aux_psijKL = par->aux_dirJKL[j]; 
		for(k = 0; k < par->K; ++k){
			//add prior and cast.
			double* psijkL = psijKL[k];
			int* aux_psijkL = aux_psijKL[k];
			double acc = 0.0;
			for(l = 0; l < data->levelsJ[j]; ++l){
				//acc += psijkL[l] = gsl_ran_gamma_int(r, 1 + aux_psijkL[l]);// +1 = uniform prior
				acc += psijkL[l] = gsl_ran_gamma(r, double(1 + aux_psijkL[l]), 1.0);// +1 = uniform prior
			}
			acc = 1.0 / acc;
			//normalize
			for (l = 0; l < data->levelsJ[j]; ++l){
				psijkL[l] *= acc;
			}
		}
	}
}
inline
void CLcm_EdImp::sam_z(){
	//THIS IS THE CRITICAL FUNCTION IN THIS SAMPLER. for K=50 the smapler spends 50% of the time executing this function. 
	// any little optimization in the inner loop has a huge impact.
	//Here we also count how many individuals belong to each class for i <= n.
	int i, j, k;
	//static int K = par->K;
	//static double probs[1000];
	std::vector<double> probs(par->K); //reallocate every time, but not such a big deal.
	//CVariable_Container& psiJKL_var = (*par)["psiJKL"];
	//memset(par->countK, 0, sizeof(int)*par->K);// <- Now this is cleared in the X2 section. MAKE SURE OF INITIALIZING IT TO ZEROS.
	std::fill(par->count1K, par->count1K + par->K, 0);
	//Z1
	for (i = 0; i < data->n; ++i){
		std::copy(par->nuK, par->nuK + par->K, &(probs[0]));
		int *  restrict xiJ = par->xIJ[i];
		for (j = 0; j < par->J; ++j){
			int l = xiJ[j];
			//for some reason the new code is faster with g++
			double ** restrict pjKL = par->psiJKL[j];
			//for (k = 0; k < par->K; k++){
			//	probs[k] *= pjKL[k][xi];
			for (double * restrict p = &(probs[0]); p != &(probs[0]) + par->K; ++p, ++pjKL){
				*p *= (*pjKL)[l];
			}
		}
		//sample zi
		par->zI[i] = k = dan_multinomial_1(r, par->K, &(probs[0]), false);
		par->count1K[k]++;
	}
}
inline
void CLcm_EdImp::sam_Z2_X2(){
	//updates:
	//-number of elements in each partition (count_partition)
	//-all individuals' responses for i > n (X2)
	//-all individuals' membership for i >n (Z2)
	//-Number of individuals in each class for i>n (countK)
	int c, k, j;
	static std::vector<double> prob(par->K);
	gsl_ran_discrete_t *d;
	//1) sample total number of individuals in each partition.
	gsl_ran_multinomial(r, data->nZeroMC, par->Nmis, par->pZeroMC_I, par->count_partition);
	//2)sample individuals within each partition
	int index = 0; //individual counter for X2.
	std::fill(par->count0K, par->count0K + par->K, 0);
	for (c = 0; c < data->nZeroMC; ++c){
		if (index == par->Nmis) return; //If we've finished with par->Nmis elements, just move on.
		//If the partition is empty, continue with the next!
		if (par->count_partition[c] == 0) continue;

		//compute the *unnormalized posterior probability of being in class k given
		// that individual is in partition c.
		std::copy(par->nuK, par->nuK + par->K, &(prob[0]));
		for (j = 0; j < data->J; ++j){
			int l = data->ZeroMC_IJ[c][j];
			for (k = 0; k < par->K; ++k){
				if (l != -1) prob[k] *= par->psiJKL[j][k][l];
			}
		}
		if (par->count_partition[c] > 1){
			//setup a sampler for the partition's zs:
			d = gsl_ran_discrete_preproc(par->K, &(prob[0]));
			for (unsigned int i = 0; i < par->count_partition[c]; ++i, ++index){
				//sample class
				par->z2_Nmax[index] = k = gsl_ran_discrete(r, d);
				++(par->count0K[k]); //increment the class counter.
				for (j = 0; j < data->J; ++j){
					par->x2_NMax_J[index][j] = (data->ZeroMC_IJ[c][j] == -1) ? 
						dan_multinomial_1_norm(r, data->levelsJ[j], par->psiJKL[j][k], 1.0) : 
						data->ZeroMC_IJ[c][j];
				}
			}
			gsl_ran_discrete_free(d);
		} else {
			par->z2_Nmax[index] = k = dan_multinomial_1(r, par->K, &(prob[0]), false);
			++(par->count0K[k]); //increment the class counter.
			for (j = 0; j < data->J; ++j){
				par->x2_NMax_J[index][j] = (data->ZeroMC_IJ[c][j] == -1) ? 
					dan_multinomial_1_norm(r, data->levelsJ[j], par->psiJKL[j][k], 1.0) : 
					data->ZeroMC_IJ[c][j];
			}
			index++;
		}
	}
}

inline
void CLcm_EdImp::sam_nu(){
	int k;
	double b = 0.0, a = 0.0, lgamma1, lgamma2, lsumgamma;
	int n_acc = 0;
	double l_acc_prod = 0.0;
	for (k = 0; k < par->K - 1; ++k){
		n_acc += par->count0K[k] + par->count1K[k];
		a = double(1 + par->count0K[k] + par->count1K[k]);
		b = par->alpha + double(par->Nmis + data->n - n_acc);
		lgamma1 = dan_log_gamma_1(r, a);
		lgamma2 = dan_log_gamma_1(r, b);
		lsumgamma = dan_log_sum(lgamma1, lgamma2);
		par->log_nuK[k] = (lgamma1 - lsumgamma + l_acc_prod);
		l_acc_prod += lgamma2 - lsumgamma; //log(1-V)
		par->nuK[k] = exp(par->log_nuK[k]);
	}
	par->log_nuK[par->K -1] = l_acc_prod;
	par->nuK[par->K - 1] = exp(par->log_nuK[par->K-1]);

}

inline
void CLcm_EdImp::sam_alpha(){
	par->alpha = gsl_ran_gamma(r, par->a_alpha + double(par->K) - 1.0, 1.0 / (par->b_alpha - par->log_nuK[par->K-1]));
}


typedef std::pair<double, int> _CLcm_EdImp_mypair;
bool _CLcm_EdImp_comparator (const _CLcm_EdImp_mypair& l, const _CLcm_EdImp_mypair& r){
	return (l.first < r.first);
}
void CLcm_EdImp::Initializes(){
	double g1, g2, gs;
	CChain::Initializes();
	std::fill(par->count0K, par->count0K + par->K,0);
	std::fill(par->count1K, par->count1K + par->K,0);
	par->alpha = 1.0;
	double lacc = 0.0;
	for (int k = 0; k < par->K - 1; k++){
		g1 = dan_log_gamma_1(r, 1);
		g2 = dan_log_gamma_1(r, 1);
		gs = dan_log_sum(g1, g2);
		par->log_nuK[k] = g1 - gs + lacc;
		lacc += g2 - gs;
	}
	par->log_nuK[par->K -1] = lacc;
	par->nuK[par->K-1] = exp(lacc);
	std::vector<std::vector<int> > v(par->J);
	std::vector<int> denom(par->J);
	int den;
	for (int j =0; j < par->J; j++){
		den = 1;
		for( int l = 0; l < par->levelsJ[j]; l++) v[j].push_back(1);
		for (int i = 0; i < par->n; i++){
			if (par->xIJ[i][j] != -1){
				++v[j][ par->xIJ[i][j] ];
				++den;
			}
		}
		for (int l = 0; l < par->levelsJ[j]; l ++) 
			for (int k = 0; k < par->K; k++)
				par->psiJKL[j][k][l] = double(v[j][l]) / double(den);
	}
	par->alpha = 0.3 / double(par->K);
	for(int k =0; k < 4; k++){
		par->nuK[k] = 0.25 * 0.9;
	}
	for(int k = 4; k < par->K; k++){
		par->nuK[k] = 0.1 / double (par->K -4);
	}
	if (this->verbose){
		DAN_PRINTF("WARMING UP...\n");
	}
	for (int i = 0; i < 500; ++i){
		sam_z();
		sam_psi();
		sam_nu();
		//truncation!
		compute_probs_miss();
		sam_Nmis();
		sam_Z2_X2();
		par->compute_k_star();
		//and regularization!
		//sam_alpha();
		if(this->verbose){
			DAN_PRINTF("\r iter = %d kstar = %d \n", i, par->k_star);
		}
	}
	//reorder decreasingly according to nuK. 
	std::vector<_CLcm_EdImp_mypair> s_index(par->K);
	for (int k = 0; k < par->K; k++){
		s_index[k].first = 1.0 - par->nuK[k];
		s_index[k].second = k;
	}
	std::sort(s_index.begin(), s_index.end(), _CLcm_EdImp_comparator);
	//CParams_Lcm_EdImp* tmp = new CParams_Lcm_EdImp(data, par->K, par->N_mis_max, par->a_alpha, 
	//	par->b_alpha, par->a_error, par->b_error,par->eMaskIJ);
	CParams_Lcm_EdImp* tmp = new CParams_Lcm_EdImp(*par);
	for (int k = 0; k < par->K; ++k){
		int indx = s_index[k].second;
		tmp->nuK[k] = par->nuK[indx];
		tmp->log_nuK[k] = par->log_nuK[indx];
		//tmp->countK[k] = par->countK[indx];
		tmp->count0K[k] = par->count0K[indx];
		tmp->count1K[k] = par->count1K[indx];
	}
	for (int k = 0; k < par->K; ++k){
		par->nuK[k] = tmp->nuK[k];
		par->log_nuK[k] = tmp->log_nuK[k];
		//par->countK[k] = tmp->countK[k];
		par->count0K[k] = tmp->count0K[k];
		par->count1K[k] = tmp->count1K[k];
	}
	for (int j = 0; j < par->J; j++){
		for (int k = 0; k < par->K; ++k){
			std::copy(par->psiJKL[j][ s_index[k].second ], par->psiJKL[j][ s_index[k].second ] + par->levelsJ[j], tmp->psiJKL[j][k]);
		}
	} 
	for (int j = 0; j < par->J; j++){
		for (int k = 0; k < par->K; ++k){
			std::copy(tmp->psiJKL[j][k], tmp->psiJKL[j][k] + par->levelsJ[j], par->psiJKL[j][k]);
		}
	} 
	for (int k = 0; k < par->K; ++k){
		for (int i = 0; i < data->n; ++i){
			if (k == s_index[k].second) tmp->zI[i] = k;
		}
		for (int i = 0; i < par->Nmis; ++i){
			if (k == s_index[k].second) tmp->z2_Nmax[i] = k;
		}
	}
	for (int i = 0; i < data->n; ++i){
			par->zI[i] = tmp->zI[i];
		}
	for (int i = 0; i < par->Nmis; ++i){
		par->z2_Nmax[i] = tmp->z2_Nmax[i];
	}
	delete tmp;
	//Don't know why
	//std::fill(par->count0K, par->count0K +par->K, 0);
	//std::fill(par->count1K, par->count1K +par->K, 0);
}



inline
void CLcm_EdImp::compute_probs_miss(){
	//computes the probability of each partition and the aggregate
	int c, k, j;
	std::fill_n(par->pZeroMC_I, par->nZeroMC, 0.0);
	par->prob_zero = 0.0;
	for (c = 0; c < data->nZeroMC; ++c){
		for (k = 0; k < par->K; ++k){
			double prod = par->nuK[k];
			for (j = 0; j < data->J; ++j){
				int l = data->ZeroMC_IJ[c][j];
				if (l != -1) prod *= par->psiJKL[j][k][l];
			}
			par->pZeroMC_I[c] += prod;
		}
		par->prob_zero += par->pZeroMC_I[c];
	}
}

inline
void CLcm_EdImp::sam_Nmis(){
	//GSL: "This function returns a random integer from the negative binomial distribution, 
	// the number of failures occurring before n successes in independent trials with probability 
	// p of success. The probability distribution for negative binomial variates is,
	// p(k) = {\Gamma(n + k) \over \Gamma(k+1) \Gamma(n) } p^n (1-p)^k NOTE THE +1 IN GAMMA(K+1)!!!!

	//int failsafe = 0;
	//do{
	//	par->Nmis = gsl_ran_negative_binomial(r, 1.0 - par->prob_zero, data->n); //<-=-this one is the correct.
	//	if (par->Nmis > par->N_mis_max) DAN_ERR_NOABORT("Nmis sampled beyond Nmis_max");
	//	if (failsafe++ > 10000){
	//		DAN_ERR_EXIT("More than 10000 attempts to sample from Nmis. Aborting");
	//	}
	//} while (par->Nmis > par->N_mis_max); //
	
	par->Nmis = gsl_ran_negative_binomial(r, 1.0 - par->prob_zero, data->n); //<-=-this one is the correct.
	if (par->Nmis > par->N_mis_max){
		try {
			par->resize_Nmis_max(ceil(par->Nmis/0.8) + 1);
		} catch (const std::exception& e){
			DAN_ERR_EXIT(e.what());
		}
		DAN_ERR_NOABORT("N_mis_max increased to %d", par->N_mis_max);
	}
	//This corresponds to an (improper) prior P(Nmis|n) \propto n/(Nmis+n). This ensures that after integrating Nmis
	// the joint probability correspond to the correct truncated distribution.
}


///////////////////////////////////////
//////////////////////////////////////
inline
void CLcm_EdImp::sam_e_x_Gibbs(){
	//imputation step as gibbs sampler.
	//std::fill(par->e_marginal_J, par->e_marginal_J + par->J, 0); //clear marginal frequcies of error.
	par->e_marginal = 0;
	//shuffle permutations
	for (int i = 0; i < data->n; i ++){
		gsl_ran_shuffle(r, par->permutationJ, par->J, sizeof(int));
		//first determine if we have to do anything for this datapoint.
		if (!par->is_mask_unknown(i)) continue; //if no unknown mask, just try the next data point.
												//The error indicator must be in 0 from the intial value and,
												//the contribution to the marginal frequency should also be 0.
		for (int jj = 0; jj < par->J; jj++){
			int n_A = 0;
			int nA_minus_Y = 0;
			int j = par->permutationJ[jj]; //take the current working coordinate from the random permutation.
			if (par->eMaskIJ[i][j] == 0) continue; //if eMask is 0, then we should not change x_ij nor eIJ (and the contribution to the margin is also 0).
			get_valid_levels_for_j(par->xIJ[i], data->ZeroMC_IJ, j, data->levelsJ[j], data->J, data->nZeroMC, _A, n_A);// get set A_ij
			int k = par->zI[i];
			//1) Sampling eIJ
			int Yij = data->x[i][j]; //observed value
			double *psi_jkL = par->psiJKL[j][k];
			bool _Y_in_A = false;
			for (int ind = 0; ind < n_A; ++ind){
				if (_A[ind] != Yij) {
					_A_minus_Y[ nA_minus_Y++ ] = _A[ind] ; 
				} else {
					_Y_in_A = true;		
				}
			}
			if (n_A == 1){
				//special case. There's only one possibility.
				par->xIJ[i][j] = _A[0];
				par->e_marginal += par->eIJ[i][j] = _A[0] == Yij ? 0 : 1; 
				continue;
			} 
			for (int ind = 0; ind < nA_minus_Y; ind++){
				_probs[ind] = psi_jkL[ _A_minus_Y[ind] ];
			}
			if (!_Y_in_A){
				par->eIJ[i][j] = 1;
			} else {
				//at this point we have to sample eIJ.
				double p[2]; //<--- ADD STATIC QUALIFIER
				//p[0] = (1.0 - par->epsilonJ[j]) * psi_jkL[Yij];
				p[0] = (1.0 - par->epsilon) * psi_jkL[Yij];
				p[1] = 0.0;
				for (int ind = 0; ind < nA_minus_Y; ind++){
					p[1] += _probs[ind];//  = psi_jkL[ _A_minus_Y[ind] ];
				}
				//p[1] *= par->epsilonJ[j] * par->qJ[j];
				p[1] *= par->epsilon * par->qJ[j]; 
				par->eIJ[i][j] = gsl_rng_uniform_pos(r) * (p[0] + p[1]) > p[0] ? 1 : 0;
			}

			//2) Now sample X conditional on e
			if (par->eIJ[i][j] == 0){
				par->xIJ[i][j] = Yij;
			} else if (nA_minus_Y == 1) {
				par->xIJ[i][j] = _A_minus_Y[0]; 
			} else {
				par->xIJ[i][j] = _A_minus_Y[ dan_multinomial_1(r, nA_minus_Y, _probs, false) ];
			}
			//par->e_marginal_J[j] += par->eIJ[i][j]; //accumulate
			par->e_marginal += par->eIJ[i][j]; //accumulate
		}
	}
}

