/* 
 * Copyright (C) 2007-2019 Daniel Manrique-Vallier
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <string.h>
#include <math.h>
#include <set>
#include <numeric>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf.h> //special functions (for gamma)

#include "daniel2/dan_math_gsl.h"
#include "CData_DM_MCZ.h"
#include "CLcm_Imputation.h"
#include "margin_conditions.h"
#include "definitions.h"

//--------------------------------------------------------------------------------
// Implementation of class CParams_Lcm_Imputation
//--------------------------------------------------------------------------------
void CParams_Lcm_Imputation::class_construct(int _Nmis_max, int** _MCZ, int _nZeroMC, int **_x){
	//this assumes that base class has been initialized
	//initialize members.
	//this->Nmis = _Nmis;
	this->N_mis_max = _Nmis_max;
	this->nZeroMC = _nZeroMC;

	//allocate extra.
	_add_parameter("MCZ", CPar_Data_Type::T_INT, &MCZ, 2, _nZeroMC, J); 
	_add_parameter("xIJ", CPar_Data_Type::T_INT, &xIJ, 2, n, J); 
	_add_parameter("pZeroMC_I", CPar_Data_Type::T_DOUBLE, &pZeroMC_I, 1, _nZeroMC);
	_add_parameter("z2_Nmax", CPar_Data_Type::T_INT, &z2_Nmax, 1, _Nmis_max);
	_add_parameter("x2_NMax_J", CPar_Data_Type::T_INT, &x2_NMax_J, 2, _Nmis_max, J);
	_add_parameter("count_partition", CPar_Data_Type::T_INT, &count_partition, 1, _nZeroMC);
	_add_parameter("permutationJ", CPar_Data_Type::T_INT, &permutationJ, 1, J);
	//register existing.
	_add_existing_scalar("prob_zero", CPar_Data_Type::T_DOUBLE, &prob_zero);
	_add_existing_scalar("Nmis", CPar_Data_Type::T_INT, &Nmis);
	_add_existing_scalar("N_mis_max", CPar_Data_Type::T_INT, &N_mis_max);
	_add_existing_scalar("nZeroMC", CPar_Data_Type::T_INT, &nZeroMC);
	//fill x with data
	(*this)["xIJ"].copy_from_raw(_x[0]); //NOTE THAT HERE X INCLUDES INCONSISTENT VALUES!!! HAVE TO INITIALIZE.
	//copy MCZ
	(*this)["MCZ"].copy_from_raw(_MCZ[0]);
	//initial permutation
	for (int j = 0; j < J; j++) permutationJ[j] = j;
	//temporary storage:
	_add_parameter("tmp_doubleJ", CPar_Data_Type::T_DOUBLE, &tmp_doubleJ, 1, J);
	_add_parameter("tmp_intJ", CPar_Data_Type::T_INT, &tmp_intJ, 1, J);
}

void CParams_Lcm_Imputation::initizalize(gsl_rng *r){
	std::vector<double> p(this->L); //reserve at least the maximum number of levels.
	std::fill(p.begin(), p.end(), 1.0); //this is a vector of unnormalized uniform probabilities.
	CParams_Lcm_Bay_Reg::initizalize(r); //parent class
	prob_zero = 0.0;
	Nmis = 0;
	for (int i = 0; i < this->nZeroMC; i++){
		count_partition[i] = 0;
	}
	(*this)["count0K"].fill<int>(0);
	(*this)["count1K"].fill<int>(0);

	//Initialize values for xIJ. (whatever. The only requirement is that they're not inconsistent.
	bool badvalue;
	for(int i = 0; i < n; ++i){
		std::vector<int> x_working(xIJ[i], xIJ[i]+J);
		do{
			for (int j = 0; j < J; ++j){
				if (xIJ[i][j] == -1) {
					x_working[j] = dan_multinomial_1(r,  this->levelsJ[j], &(p[0]), false);
				}
			}
			//check
			badvalue = false;
			for (int ii  = 0; ii < this->nZeroMC; ++ii){
				if (!check_x_notin_mu(x_working.begin(), x_working.end(), MCZ[ii])) {
					badvalue = true; 
					break; 
				}
			}
		} while ( badvalue );
		std::copy<std::vector<int>::iterator, int*>(x_working.begin(), x_working.end(), xIJ[i]);
	}
}


//--------------------------------------------------------------------------------
// Implementation of class CLcm_Imputation
//--------------------------------------------------------------------------------


void CLcm_Imputation::Update() {

	//regular stuff
	sam_z();
	sam_psi();
	sam_nu();
	//truncation!
	compute_probs_miss();
	sam_Nmis();
	sam_Z2_X2();
	//and regularization!
	sam_alpha();
	// Imputation step
	sam_x();
	par->compute_k_star();
}

// Samplers for the full conditional distributions.
//  each function update the values on the *par object directly.
inline 
void CLcm_Imputation::sam_x(){
	//TODO: <--
	// [ ] implement switch clause for different methods.
	// [ ] pre-fetch missing data at initialization stage

	// 1) REJECTION SAMPLING
	for (int i = 0; i < par->n; i++){
		//int failsafe = 0;
		bool reject;
		do {
			reject = false;
			bool imputed = false;
			for (int j = 0; j < par->J; j++){
				if (data->x[i][j] != -1) continue; //next j
				par->xIJ[i][j] = dan_multinomial_1(this->r, par->levelsJ[j], par->psiJKL[j][ par->zI[i] ], true);
				imputed = true;
			}
			if (!imputed) break; //next i because reject = F
			// For now we just keep trying until eternity.
			//if (failsafe++ > 50000){
			//		DAN_ERR_EXIT("more than 50000 rejections in imputation. Aborting\n");
			//}
			for (int m = 0; m < par->nZeroMC; m++){
				reject = !check_x_notin_mu(par->xIJ[i], par->xIJ[i] + par->J, data->ZeroMC_IJ[m]);
				if (reject) break;
			}
		} while (reject);
	}
}


inline
void CLcm_Imputation::sam_psi(){
	//static double dirL[1000];
	int i, j, k, l;
	std::fill_n(par->aux_dirJKL[0][0], data->J * par->K *data->L, 0);
	//the data part.
	for (i = 0; i < data->n; ++i){
		int k = par->zI[i];
		for (j = 0; j < par->J; ++j){
			++par->aux_dirJKL[j][k][ par->xIJ[i][j] ];
		}
	}
	//the imputed part.
	for (i = 0; i < par->Nmis; ++i){
		k = par->z2_Nmax[i];
		for (j = 0; j < par->J; ++j){
			++par->aux_dirJKL[j][k][ par->x2_NMax_J[i][j] ];
		}
	}
	
	//add the prior (+1) and sample
	for(j = 0; j < data->J; ++j){
		double **psijKL = par->psiJKL[j]; 
		int **aux_psijKL = par->aux_dirJKL[j]; 
		for(k = 0; k < par->K; ++k){
			//add prior and cast.
			double* psijkL = psijKL[k];
			int* aux_psijkL = aux_psijKL[k];
			double acc = 0.0;
			for(l = 0; l < data->levelsJ[j]; ++l){
				//acc += psijkL[l] = gsl_ran_gamma_int(r, 1 + aux_psijkL[l]);// +1 = uniform prior
				acc += psijkL[l] = gsl_ran_gamma(r, double(1 + aux_psijkL[l]), 1.0);// +1 = uniform prior
			}
			acc = 1.0 / acc;
			//normalize
			for (l = 0; l < data->levelsJ[j]; ++l){
				psijkL[l] *= acc;
			}
		}
	}
}
inline
void CLcm_Imputation::sam_z(){
	//THIS IS THE CRITICAL FUNCTION IN THIS SAMPLER. for K=50 the smapler spends 50% of the time executing this function. 
	// any little optimization in the inner loop has a huge impact.
	//Here we also count how many individuals belong to each class for i <= n.
	int i, j, k;
	//static int K = par->K;
	static double probs[1000];
	//memset(par->countK, 0, sizeof(int)*par->K);// <- Now this is cleared in the X2 section. MAKE SURE OF INITIALIZING IT TO ZEROS.
	std::fill(par->count1K, par->count1K + par->K, 0);
	//Z1
	for (i = 0; i < data->n; ++i){
		std::copy(par->nuK, par->nuK + par->K, probs);
		int *  restrict xiJ = par->xIJ[i];
		for (j = 0; j < par->J; ++j){
			int l = xiJ[j];
			//for some reason the new code is faster with g++
			double ** restrict pjKL = par->psiJKL[j];
			//for (k = 0; k < par->K; k++){
			//	probs[k] *= pjKL[k][xi];
			for (double * restrict p = probs; p != probs + par->K; ++p, ++pjKL){
				*p *= (*pjKL)[l];
			}
		}
		//sample zi
		par->zI[i] = k = dan_multinomial_1(r, par->K, probs, false);
		par->count1K[k]++;
	}
}
inline
void CLcm_Imputation::sam_Z2_X2(){
	//updates:
	//-number of elements in each partition (count_partition)
	//-all individuals' responses for i > n (X2)
	//-all individuals' membership for i >n (Z2)
	//-Number of individuals in each class for i>n (countK)
	int c, k, j;
	static double prob[1000];
	gsl_ran_discrete_t *d;
	//1) sample total number of individuals in each partition.
	gsl_ran_multinomial(r, data->nZeroMC, par->Nmis, par->pZeroMC_I, par->count_partition);
	//2)sample individuals within each partition
	int index = 0; //individual counter for X2.
	std::fill(par->count0K, par->count0K + par->K, 0);
	for (c = 0; c < data->nZeroMC; ++c){
		//compute the *unnormalized posterior probability of being in class k given
		// that individual is in partition c.
		std::copy(par->nuK, par->nuK + par->K, prob);
		for (j = 0; j < data->J; ++j){
			int l = data->ZeroMC_IJ[c][j];
			for (k = 0; k < par->K; ++k){
				if (l != -1) prob[k] *= par->psiJKL[j][k][l];
			}
		}
		//setup a sampler for the partition's zs:
		d = gsl_ran_discrete_preproc(par->K, prob);
		for (unsigned int i = 0; i < par->count_partition[c]; ++i, ++index){
			//sample class
			par->z2_Nmax[index] = k = gsl_ran_discrete(r, d);
			++(par->count0K[k]); //increment the class counter.
			for (j = 0; j < data->J; ++j){
				par->x2_NMax_J[index][j] = (data->ZeroMC_IJ[c][j] == -1) ? 
					dan_multinomial_1_norm(r, data->levelsJ[j], par->psiJKL[j][k], 1.0) : 
					data->ZeroMC_IJ[c][j];
			}
		}
		gsl_ran_discrete_free(d);
	}
}

inline
void CLcm_Imputation::sam_nu(){
	int k;
	double b = 0.0, a = 0.0, lgamma1, lgamma2, lsumgamma;
	int n_acc = 0;
	double l_acc_prod = 0.0;
	for (k = 0; k < par->K - 1; ++k){
		n_acc += par->count0K[k] + par->count1K[k];
		a = double(1 + par->count0K[k] + par->count1K[k]);
		b = par->alpha + double(par->Nmis + data->n - n_acc);
		lgamma1 = dan_log_gamma_1(r, a);
		lgamma2 = dan_log_gamma_1(r, b);
		lsumgamma = dan_log_sum(lgamma1, lgamma2);
		par->log_nuK[k] = (lgamma1 - lsumgamma + l_acc_prod);
		l_acc_prod += lgamma2 - lsumgamma; //log(1-V)
		par->nuK[k] = exp(par->log_nuK[k]);
	}
	par->log_nuK[par->K -1] = l_acc_prod;
	par->nuK[par->K - 1] = exp(par->log_nuK[par->K-1]);

}

inline
void CLcm_Imputation::sam_alpha(){
	par->alpha = gsl_ran_gamma(r, par->a_alpha + double(par->K) - 1.0, 1.0 / (par->b_alpha - par->log_nuK[par->K-1]));
}


typedef std::pair<double, int> _CLcm_Imputation_mypair;
bool _CLcm_Imputation_comparator (const _CLcm_Imputation_mypair& l, const _CLcm_Imputation_mypair& r){
	return (l.first < r.first);
}
void CLcm_Imputation::Initializes(){
	double g1, g2, gs;
	CChain::Initializes();
	par->alpha = 1.0;
	double lacc = 0.0;
	std::fill(par->count0K, par->count0K + par->K,0);
	std::fill(par->count1K, par->count1K + par->K,0);
	for (int k = 0; k < par->K - 1; k++){
		g1 = dan_log_gamma_1(r, 1);
		g2 = dan_log_gamma_1(r, 1);
		gs = dan_log_sum(g1, g2);
		par->log_nuK[k] = g1 - gs + lacc;
		lacc += g2 - gs;
	}
	par->log_nuK[par->K -1] = lacc;
	par->nuK[par->K-1] = exp(lacc);
	std::vector<std::vector<int> > v(par->J);
	std::vector<int> denom(par->J);
	int den;
	for (int j =0; j < par->J; j++){
		den = 1;
		for( int l = 0; l < par->levelsJ[j]; l++) v[j].push_back(1);
		for (int i = 0; i < par->n; i++){
			if (par->xIJ[i][j] != -1){
				++v[j][ par->xIJ[i][j] ];
				++den;
			}
		}
		for (int l = 0; l < par->levelsJ[j]; l ++) 
			for (int k = 0; k < par->K; k++)
				par->psiJKL[j][k][l] = double(v[j][l]) / double(den);
	}
	par->alpha = 0.3 / double(par->K);
	for(int k =0; k < 4; k++){
		par->nuK[k] = 0.25 * 0.9;
	}
	for(int k = 4; k < par->K; k++){
		par->nuK[k] = 0.1 / double (par->K -4);
	}
	if(this->verbose){
		DAN_PRINTF("WARMING UP...\n");
	}
	for (int i = 0; i < 500; ++i){
		sam_z();
		sam_psi();
		sam_nu();
		//truncation!
		compute_probs_miss();
		sam_Nmis();
		sam_Z2_X2();
		par->compute_k_star();
		//and regularization!
		//sam_alpha();
		if(this->verbose){
			DAN_PRINTF("\r iter = %d kstar = %d \n", i, par->k_star);
		}
	}
	//reorder decreasingly according to nuK. 
	std::vector<_CLcm_Imputation_mypair> s_index(par->K);
	for (int k = 0; k < par->K; k++){
		s_index[k].first = 1.0 - par->nuK[k];
		s_index[k].second = k;
	}
	std::sort(s_index.begin(), s_index.end(), _CLcm_Imputation_comparator);
	//CParams_Lcm_Imputation* tmp = new CParams_Lcm_Imputation(data, par->K, par->N_mis_max, par->a_alpha, 
	//	par->b_alpha, par->a_error, par->b_error,par->eMaskIJ);
	CParams_Lcm_Imputation* tmp = new CParams_Lcm_Imputation(*par);
	for (int k = 0; k < par->K; ++k){
		int indx = s_index[k].second;
		tmp->nuK[k] = par->nuK[indx];
		tmp->log_nuK[k] = par->log_nuK[indx];
		//tmp->countK[k] = par->countK[indx];
		tmp->count0K[k] = par->count0K[indx];
		tmp->count1K[k] = par->count1K[indx];
	}
	for (int k = 0; k < par->K; ++k){
		par->nuK[k] = tmp->nuK[k];
		par->log_nuK[k] = tmp->log_nuK[k];
		//par->countK[k] = tmp->countK[k];
		par->count0K[k] = tmp->count0K[k];
		par->count1K[k] = tmp->count1K[k];
	}
	for (int j = 0; j < par->J; j++){
		for (int k = 0; k < par->K; ++k){
			std::copy(par->psiJKL[j][ s_index[k].second ], par->psiJKL[j][ s_index[k].second ] + par->levelsJ[j], tmp->psiJKL[j][k]);
		}
	} 
	for (int j = 0; j < par->J; j++){
		for (int k = 0; k < par->K; ++k){
			std::copy(tmp->psiJKL[j][k], tmp->psiJKL[j][k] + par->levelsJ[j], par->psiJKL[j][k]);
		}
	} 
	for (int k = 0; k < par->K; ++k){
		for (int i = 0; i < data->n; ++i){
			if (k == s_index[k].second) tmp->zI[i] = k;
		}
		for (int i = 0; i < par->Nmis; ++i){
			if (k == s_index[k].second) tmp->z2_Nmax[i] = k;
		}
	}
	for (int i = 0; i < data->n; ++i){
			par->zI[i] = tmp->zI[i];
		}
	for (int i = 0; i < par->Nmis; ++i){
		par->z2_Nmax[i] = tmp->z2_Nmax[i];
	}
	delete tmp;
	//NOT SURE WHY I WROTE THIS!!!!!
	//std::fill(par->count0K, par->count0K +par->K, 0);
	//std::fill(par->count1K, par->count1K +par->K, 0);
}



inline
void CLcm_Imputation::compute_probs_miss(){
	//computes the probability of each partition and the aggregate
	int c, k, j;
	std::fill_n(par->pZeroMC_I, par->nZeroMC, 0.0);
	par->prob_zero = 0.0;
	for (c = 0; c < data->nZeroMC; ++c){
		for (k = 0; k < par->K; ++k){
			double prod = par->nuK[k];
			for (j = 0; j < data->J; ++j){
				int l = data->ZeroMC_IJ[c][j];
				if (l != -1) prod *= par->psiJKL[j][k][l];
			}
			par->pZeroMC_I[c] += prod;
		}
		par->prob_zero += par->pZeroMC_I[c];
	}
}

inline
void CLcm_Imputation::sam_Nmis(){
	//GSL: "This function returns a random integer from the negative binomial distribution, 
	// the number of failures occurring before n successes in independent trials with probability 
	// p of success. The probability distribution for negative binomial variates is,
	// p(k) = {\Gamma(n + k) \over \Gamma(k+1) \Gamma(n) } p^n (1-p)^k NOTE THE +1 IN GAMMA(K+1)!!!!
	//Truncate N. JUST IN CASE. THIS SHOULDN'T HAPPEN.
	do{
		par->Nmis = gsl_ran_negative_binomial(r, 1.0 - par->prob_zero, data->n); //<-=-this one is the correct.
	} while (par->Nmis > par->N_mis_max);
	//This corresponds to an (improper) prior P(Nmis|n) \propto n/(Nmis+n). This ensures that after integrating Nmis
	// the joint probability correspond to the correct truncated distribution.
}


