/* 
 * Copyright (C) 2007-2019 Daniel Manrique-Vallier
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include<set>
#include<algorithm>
#include<vector>
#include<list>
#include<iostream>
#include "daniel2/dan_array_utils.h"
#include "margin_conditions.h"
#include "definitions.h"
#include <math.h>

template<typename T_Vector>
inline
void get_relevant_cols(const T_Vector &condition, std::vector<int> &lista, const std::vector<int> &irrelevant, const int &length){
	std::set<int> EX(irrelevant.begin(), irrelevant.end());
	lista.clear();
	for (int i = 0; i < length; i++){
		if (condition[i] != -1 && EX.count(i) ==0){
			lista.push_back(i);
		}
	}
}

template<typename T>
inline
int next_arr(T dat, const std::vector<int> &seq, const int* &levelsJ, const int &order = 0){
	int j = order;
	int nCol = seq.size();
	bool carry;
	do{
		carry = false;
		dat[seq[j]]++;
		if(dat[seq[j]] >= levelsJ[seq[j]]) { 
			dat[seq[j]] = 0; 
			j ++;
			carry =true;
		}
	} while(carry && j < nCol);
	return (j);
}

template<typename T_Container>
void get_seq_exploration(const T_Container &excl_vec, std::vector<int> &seq, const int &J, const std::vector<int> &irrelevant){
	std::set<int> already(irrelevant.begin(), irrelevant.end());
	seq.clear();
	//int n = excl_vec.size();
	for (typename T_Container::const_iterator it = excl_vec.begin(); it != excl_vec.end(); it++){
		for (int j = 0; j < J; j++){
			if (((*it)[j]!= -1) && (already.count(j) == 0)){
				already.insert(j);
				seq.push_back(j);
			}
		}
	}
} 

void expand_lite(std::list<std::vector<int> > &excl_list, const int* curr_cond, const int *levels, const int &J){
	std::list<std::vector<int> > compar_list;
	std::vector<int> curr_cond_v(curr_cond, curr_cond + J);
	//1) test one by one and only keep those that are not disjoint.
	for (std::list<std::vector<int> >::iterator it = excl_list.begin(); it != excl_list.end(); it++){
		if (!check_disjoint(it->begin(), it->end(), curr_cond)){
			compar_list.push_back(*it);
		}
	}
	//Now compar_list is a vector of pointers to the working conditions (vectors).
	// if it's empty, the condition can enter as is
	if (compar_list.size() == 0){
		excl_list.push_back(curr_cond_v);
		return;
	}
	std::vector<int> cols; //sequence of column expansion
	std::vector<int> fixed_cols; //fixed columns
	get_relevant_cols(curr_cond, fixed_cols, cols, J);
	get_seq_exploration(compar_list, cols, J, fixed_cols);
	//special case: if cols is empty, then 
	if (cols.size() == 0){
		excl_list.push_back(curr_cond_v);
		return;
	}
	std::vector<int> current_expanded(curr_cond, curr_cond + J);
	for(std::vector<int>::iterator it = cols.begin(); it != cols.end(); it++){
		current_expanded[*it] = 0;
	}
	//bool done = false;
	unsigned int order;
	do{
		//check that is disjoint with *every* element in the list
		int  disj = 1;
		for (std::list<std::vector<int> >::iterator it = compar_list.begin(); it != compar_list.end(); it++){
			disj *= check_disjoint(it->begin(), it->end(), current_expanded.begin());
		}
		if (disj) excl_list.push_back(current_expanded);
		order = next_arr(&(current_expanded[0]), cols, levels);
	}while(order < cols.size());
}
typedef std::pair<double, int*> margin_mypair;
bool margin_comparator (const margin_mypair& l, const margin_mypair& r){
	return (l.first < r.first); //This is the correct one
	//return (l.first > r.first); 
}

int** MC_to_MCpartition_old( int** mc, const int&J, const int &nCond, const int* levels, int &SizeDest){
	std::list<std::vector<int> > exclusive;
	//reorder according to set size.
	std::vector<margin_mypair> sizes(nCond);
	for (int i = 0; i < nCond; i++){
		sizes[i].second = mc[i];
		sizes[i].first = 1.0;
		for (int j = 0; j < J; j++){
			if (mc[i][j] != -1) sizes[i].first *= double(levels[j]);
		}
	}
	std::sort(sizes.begin(), sizes.end(), margin_comparator);
	exclusive.push_back(std::vector<int>(sizes[0].second, sizes[0].second + J));
	for (int i = 1; i < nCond; i++){
		expand_lite(exclusive, sizes[i].second, levels, J);
	}
	int C = exclusive.size();
	int **it_dest, **dest;
	//HERE!!! <-- CHANGE DATA ALLOCATION AND CHECK WHERE THIS IS FREED.
	it_dest = dest = (int**) dan_flat2array2D(dan_malloc(sizeof(int)*J*C, "", ""), sizeof(int),C, J);
	
	for(std::list<std::vector<int> >::iterator it = exclusive.begin(); it != exclusive.end(); it++){
		std::copy(it->begin(), it->end(), (it_dest++)[0]);
	}
	SizeDest = C;
	return(dest);
}

bool check_disjoint_MC(int** cond, int nCond, int nVars){
	for (int** it1 = cond; it1 != cond + nCond - 1; it1++){
		for (int** it2 = it1 + 1; it2 != cond + nCond; it2++){
			if (!check_disjoint(*it1, *it1 + nVars, *it2)){
				return(false);
			}
		}
	}
	return (true);
}

template<typename itT_A, typename itT_B>
inline bool check_condA_incl_condB(const itT_A &A_begin, const itT_A &A_end, const itT_B &B_begin){
	// A \ninc B iff a!=-1 and b!=a OR A=-1 and B!=-1
	itT_B itb = B_begin;
	for(itT_A ita = A_begin; ita != A_end; ita++, itb++){
		if((*ita != -1 && *itb != *ita) || (*ita == -1 && itb != -1)) return(false);
	}
	return(true);
}




///////////////New implementation of MCZ processing//////////////
void remove_B_from_A (std::vector<int> &A, const std::vector<int> &B, std::list<std::vector<int> >& result, const int* levels){
	//THIS FUNCTION APPENDS TO "result", it doesn't clear it!
	//check if they're already disjoint
	if (check_disjoint(A.begin(), A.end(), B.begin())){
		result.push_back(A);
		return;
	}
	//They intersect, so we'll remove the overlapping part.
	int J = A.size(); 
	//Determine the indices to expand. These are all the placeholders in A that have a fixed value in B
	std::vector<int> indices;
	for (int j = 0; j < J; j ++ ){
		if (A[j] == -1 && B[j] != -1){
			indices.push_back(j);
		}
	}
	//If no expansion necesary, B completely contains A. Result is empty!
	if (indices.size() == 0){
		return; 
	}
	//We generate each expanded vector from A sequentially. 
	// we start with the 0...0 expansion.
	std::vector<int> current_expansion(A.begin(), A.end());
	for(std::vector<int>::iterator it = indices.begin(); it != indices.end(); it++){
		current_expansion[*it] = 0;
	}
	//check each of the expanded vectors against B.
	unsigned int order;
	do{
		if(check_disjoint(B.begin(), B.end(),current_expansion.begin())){
			// if they're disjoint, we keep the current expansion.
			result.push_back(current_expansion);
		}
		//generate the next expansion in the lexicographical sequence.
		order = next_arr(&(current_expansion[0]), indices, levels);
	} while (order < indices.size());
}

void remove_B_from_collectionC(std::list<std::vector<int> > &C, const std::vector<int> & B, const int* levels){
	//removes vector B from each vector in the list C.
	// ony expands what's needed.
	std::list<std::vector<int> > tmp_list;
	for(std::list<std::vector<int> >::iterator it = C.begin(); it != C.end(); it++){
		remove_B_from_A(*it, B, tmp_list,levels);
	}
	//change this for a proper insertion.
	C.clear();
	C.insert(C.end(), tmp_list.begin(), tmp_list.end());
}

//New version Jul-2015
void MC_to_MCpartitionList(int** mc, const int& J, const int &nCond, const int* levels, std::list<std::vector<int> >& Dest){
	//Returns the result in the list of vectors "Dest". Dest is erased in this operation.

	//reorder according to set size. This way we minimize the expansions.
	std::vector<margin_mypair> sizes(nCond);
	for (int i = 0; i < nCond; i++){
		sizes[i].second = mc[i];
		sizes[i].first = 0.0;
		for (int j = 0; j < J; j++){
			if (mc[i][j] != -1) sizes[i].first += log(double(levels[j]));
		}
	}
	std::sort(sizes.begin(), sizes.end(), margin_comparator);
	Dest.clear();
	Dest.push_back(std::vector<int>(sizes[0].second, sizes[0].second + J));
	for (int i = 1; i < nCond; i++){
		std::list<std::vector<int> > curr_expand;
		std::vector<int> omega_i(sizes[i].second, sizes[i].second + J);
		curr_expand.push_back(omega_i);
		for (int k = 0; k < i; k++){
			remove_B_from_collectionC(curr_expand, std::vector<int>(sizes[k].second, sizes[k].second+J), levels);
		}
		//debug
		//std::cout << "i= "<< i << " (";
		//for (int z = 0; z < J; z++){
		//	std::cout << omega_i[z] << " ";
		//}
		//std::cout << ") expansion size " << curr_expand.size() << std::endl;
		//for (std::list<std::vector<int> >::iterator it = curr_expand.begin(); it != curr_expand.end(); it++){
		//	 std::cout << "\t";
		//	for (int z = 0; z < J; z++){
		//		std::cout << (*it)[z] << " ";
		//	}
		//	std::cout << std::endl;
		//}
		Dest.insert(Dest.end(), curr_expand.begin(), curr_expand.end());
	}
}


//Interface for compatibility with older code.
int** MC_to_MCpartition( int** mc, const int&J, const int &nCond, const int* levels, int &SizeDest){
	//wrapper for compatibility with older code. User MC_toMCpartitionList(...) instead.
	std::list<std::vector<int> > exclusive;
	MC_to_MCpartitionList(mc, J, nCond, levels, exclusive);
	//copy the result into a usable format.
	int C = exclusive.size();
	int **it_dest, **dest;
	//HERE!!! <-- CHANGE DATA ALLOCATION AND CHECK WHERE THIS IS FREED.
	it_dest = dest = (int**) dan_flat2array2D(dan_malloc(sizeof(int)*J*C, "", ""), sizeof(int),C, J);
	
	for(std::list<std::vector<int> >::iterator it = exclusive.begin(); it != exclusive.end(); it++){
		std::copy(it->begin(), it->end(), (it_dest++)[0]);
	}
	SizeDest = C;
	return(dest);
}

//#include "test_data.h"
//void test_MC_conversion(){
//	int ln[] = {4,10};
//	int data[] = {
//		 1, 1,-1,-1,-1,-1,-1,-1,-1,-1,
//		-1,-1, 1, 1,-1,-1,-1,-1,-1,-1,
//		-1,-1,-1,-1, 1, 1,-1,-1,-1,-1,
//		-1,-1,-1,-1,-1,-1, 1, 1,-1,-1,
//		-1,-1,-1,-1,-1,-1,-1,-1, 1, 1,
//	};
//	int levels[] ={10, 10, 2, 6, 5, 5, 3, 10, 2, 2};
//	int** arr = (int**)dan_flat2arrayND_ln(data, sizeof(int), 2, ln);
//	bool disj = check_disjoint_MC(arr, ln[0], ln[1]);
//	std::cout << "disjoint = " << (disj ? "yes" : "no") << std::endl ;
//	int ndest;
//	int** final = MC_to_MCpartition(arr, ln[1], ln[0], levels, ndest);
//	std::cout << "number of partitions " << ndest << std::endl;
//	disj = check_disjoint_MC(final, ndest, ln[1]);
//	std::cout << "disjoint = " << (disj ? "yes" : "no") << std::endl ;
//}

//void main(){
//	test_MC_conversion();
//}
void main_pieces(){
	int levels[] = {2,3,4,5};
	int Ac[] ={1,1,-1,-1};
	int Bc[] = {1,-1,1,-1};
	int Cc[] = {1,-1,2,1};
	const int* lev = levels;
	std::vector<int> A(Ac, Ac + 4);
	std::vector<int> B(Bc, Bc + 4);
	std::vector<int> C(Cc, Cc + 4);
	std::list<std::vector<int> > col;
	std::list<std::vector<int> > res;
	col.push_back(B);
	col.push_back(C);
	remove_B_from_collectionC(col, A, lev);
}
