marg_prob <- function(df, margin = c(1,2), N = NA){
  #Computes marginal probabilites from a data frame
  th <- prop.table(table(df[,margin]))
  th.v <- th*(1-th)/NROW(df) * (1-NROW(df)/N)
  return(list(point=th, var=th.v))
}

sam_stat <- function(df, m_way=2, N){
  tmp <- combn(x=1:NCOL(df), m=m_way, function(x)marg_prob(df,x,N),simplify=F)
  p <- unlist(sapply(tmp, function(x) as.numeric(x$point), simplify=T))
  var <- unlist(sapply(tmp, function(x) as.numeric(x$var), simplify=T))
  return(list(point=p, var=var))
}

fn_contaminate2 <- function(orig, err_rate = rep(0.1, 10)){
  #contaminate data with a uniform errors model
  #(fast version)
  subs_mis <- orig
  ssize <- NROW(orig)
  J <- NCOL(orig)
  errors <- numeric(J)
  mask <- matrix(FALSE, ncol = J, nrow = ssize)
  for (j in 1:J){
    mask[,j] <- rbinom(ssize, size = 1, err_rate[j]) == 1
    lev <- levels(orig[,j])
    nlev <- nlevels(orig[,j]) 
    for (l in 1:nlev){
      indx <- mask[,j] & (orig[,j] == lev[l])
      subs_mis[indx, j] <-  sample(lev[-l], size = sum(indx), replace = T)
    }
  }
  colnames(mask) <- names(subs_mis)
  rows_with_errors <- which(apply(subs_mis != orig, MARGIN = 1, FUN = function(x)sum(x)>0)==1)
  errors <- apply(mask, MARGIN = 2, FUN = sum)
  return(list(result = subs_mis, 
              original = orig, 
              error_rate = err_rate,
              mask = mask,
              rows_with_errors = rows_with_errors, 
              errors = errors))
}

fn_contaminate3 <- function(orig, err_rate = rep(0.1, NCOL(orig)), q = NULL){
  #contaminate a dataset with a general reporting and error model (fast version)
  # err_rate: probabilities of error per column
  # q: substitution probabilities. A list with as many elements as columns in "orig". Each element is the 
  #     probability of selection of each level of the variable. Uniform model when each level has the same probability within
  #     a single variable.
  #browser()
  rs <- .Random.seed
  require(dplyr)
  J <- NCOL(orig)
  L <- sapply(orig[1,], FUN = nlevels)
  n <- NROW(orig)
  if (is.null(q)){
    q <- list()
    for (j in 1:J){
      q[[j]] <- rep(1, L[j]) / L[j]
    }
  }
  mask <- sapply(err_rate, FUN = function(x) rbinom(n, size = 1, p = x) == 1, simplify = T)
  contaminated <- orig
  samp.matrix <- orig %>% 
    mutate_each(funs(as.numeric(.))) %>%
    as.matrix() #Take the numerical values of the factors.
  for (j in 1:J){
    t_mat <- matrix(rep(q[[j]], L[j]), nrow =L[j], byrow=T) * (1-diag(1, L[j])) #transition matrix for contaminated entries. Diag = 0
    for (l in 1:L[j]){
      indx <- mask[,j] & (samp.matrix[,j] == l)
      samp.matrix[indx, j] <- sample(1:(L[j]), size = sum(indx), prob = t_mat[l,], replace = T)
    }
    contaminated[,j] <- factor(samp.matrix[,j], levels = 1:(L[j]), labels = levels(orig[1,j]))
  }
  rows_with_errors <- which(apply(contaminated != orig, MARGIN = 1, FUN = function(x)sum(x)>0)==1)
  errors <- apply(mask, MARGIN = 2, FUN = sum)
  res <- list(result = contaminated, 
              original = orig, 
              error_rate = err_rate,
              mask = mask,
              q = q,
              rows_with_errors = rows_with_errors, 
              errors = errors)
  attr(res, "seed") <- rs
  attr(res, 'date') <- date()
  return(res)
}


fn_blank_mask  <- function(dataset, mask){
  #replace NAs wherever mask==TRUE
  subs_FH <-dataset
  for ( i in 1:NROW(dataset)){
    for (j in 1:NCOL(dataset)){
      if (mask[i,j]){
        subs_FH[i,j] <- NA
      }
    }
  }
  return(subs_FH)
}

fn_poke_holes <- function(dataset, rate = 0.3){
  #randomly blank entries with NA.
  subs_FH <-dataset
  for ( i in 1:NROW(dataset)){
    for (j in 1:NCOL(dataset)){
      if (rbinom(n=1, size=1, prob = rate)>0) subs_FH[i,j] <- NA
    }
  }
  return(subs_FH)
}

scatter_comparison <- function(x = actual, 
                               y = FH$s.imputed, 
                               m = max(x, y), 
                               main = 'Two-way margin proportions\n (noisy sample)',
                               xlab = 'Population',
                               ylab = 'Noisy Sample'){
  plot(x = c(0,m), y = c(0,m), type = 'l' , lty = 2,
       main = main, 
       xlab = xlab,
       ylab = ylab)
  points(x, y)
}


###
#Things for fellegi-holt
fn_Create_Edit_Rules <- function(Zero_Definitions){
  #Create edit rules from MCZ definitions
  require('editrules')
  #a) domain
  lst_defs <- list()
  for (var in colnames(Zero_Definitions)){
    levs <- levels(Zero_Definitions[1,var])
    lst_defs[var] <- paste(var, ' %in% c(', paste(paste("'", levs, "'", sep = ''), collapse = ', '), ')', sep = '')
  }
  
  #b) edit rules (assumes pairwise stuff)
  
  lst_edits <- list()
  for (i in 1:NROW(Zero_Definitions)){
    tmp <- Zero_Definitions[i,]
    indx <- which(!is.na(tmp))
    nms <- names(tmp)[indx]
    lst_edits[i] <- paste('if (', nms[1], " == '", as.character(tmp[,indx[1] ]),"') ", nms[2], " != '", as.character(tmp[ ,indx[2] ]), "'", sep ='')
  }
  names(lst_defs) <- NULL
  E <- do.call("editarray", list(unlist(append(lst_defs, lst_edits))))
  return(edits=E)
}
